/* Store data for a polygon mesh: vertexes in space, polygons, attributes
from polygons.

Vertexes: are locations in space (have 3 components: X, Y, Z)
Polygons: define index vertexes that compose one polygon
Attributes: define if polygon is a portal and your texture ID

WARNING:
- after change mesh data call CMesh::mBeginAddMesh and imediatly before
  call CMesh::mEndAddMesh.
*/



#include <string.h>
#include <math.h>

#include "defs.h"

#include "node.h"

#include "nodeint.h"

#include "nodever.h"

#include "nodepol.h"

#include "nodepor.h"

#include "list.h"

#include "map.h"

#include "mesh.h"



/* constructor - init empty sector
**********************************************************/
CMesh::CMesh ()
{
}



/* destructor - desalocate data, if alocated
**********************************************************/
CMesh::~CMesh ()
{
   mDestroy ();
}



/* Prepare mesh to reseve data: disable direct access feature.
return TRUE - ok / FALSE - error
Call after change mesh data, before call CMesh::mEndAddMesh
**********************************************************/
BOOL CMesh::mBeginAddMesh ()
{
   m_oPolygons.mDisableDirectAccess ();
   m_oVertexes.mDisableDirectAccess ();

   return TRUE;
}



/* Update mesh data acording polygons and vertexes definition:
recompute polygons normals and enable direct access to data lists.
return TRUE - ok / FALSE - error
Call after define all data and before use this data.
**********************************************************/
BOOL CMesh::mEndAddMesh ()
{
   // try enable direct access feature (speed up random access)
   if ( ! m_oPolygons.mEnableDirectAccess () ||
        ! m_oVertexes.mEnableDirectAccess () )
      return FALSE;   // if error ocurr return fail

   // compute polygons plan equation (used in collision detection)
   // first compute polygon normal vector and then the D component.
   // Plan equation: AX + BY + CZ + D = 0,   where
   // (A, B, C) is a plan normal and unitary vector
   // D = AX' + BY' + CZ',   (X', Y', Z') is a point inside plan
   // (X, Y, Z) is a point to test
   m_oPolygons.mSelectStart ();
   while ( m_oPolygons.mRetExistSelection () ) {

      // the current polygon
      CNodePolygon
         *poPol = (CNodePolygon*) m_oPolygons.mRetSelected ();
      
      // define number of vertexes in polygon
      DWORD
         dwNumVertexes = poPol->m_stPol.dwNumVertexes;

      // verify if enought vertexes are defined in mesh
      if ( dwNumVertexes < 3 )
         ABORT ( MSG_ERROR_LOW_VERTEXES_IN_MESH, __FILE__, __LINE__ )

      // define 3 vertexes inside polygon plan
      CNodeVertex 
         *poVert0 = (CNodeVertex*) m_oVertexes.mRetNode ( poPol->m_stPol.padwVertexes [ 0 ] ),
         *poVert1 = (CNodeVertex*) m_oVertexes.mRetNode ( poPol->m_stPol.padwVertexes [ 1 ] ),
         *poVert2 = (CNodeVertex*) m_oVertexes.mRetNode ( poPol->m_stPol.padwVertexes [ 2 ] );
      
      // define 2 vectors inside the poly plan
      tagXYZ
         stVector1,
         stVector2;
      stVector1.nX = poVert0->m_stVertex.nX - poVert1->m_stVertex.nX;
      stVector1.nY = poVert0->m_stVertex.nY - poVert1->m_stVertex.nY;
      stVector1.nZ = poVert0->m_stVertex.nZ - poVert1->m_stVertex.nZ;
      stVector2.nX = poVert0->m_stVertex.nX - poVert2->m_stVertex.nX;
      stVector2.nY = poVert0->m_stVertex.nY - poVert2->m_stVertex.nY;
      stVector2.nZ = poVert0->m_stVertex.nZ - poVert2->m_stVertex.nZ;
      
      // compute normal vector
      tagfXYZ
         stNormal;
      stNormal.fX = (float) ( stVector1.nY * stVector2.nZ - stVector1.nZ * stVector2.nY );
      stNormal.fY = (float) ( stVector1.nZ * stVector2.nX - stVector1.nX * stVector2.nZ );
      stNormal.fZ = (float) ( stVector1.nX * stVector2.nY - stVector1.nY * stVector2.nX );
      // normalize vector (make length 1)
      float
         fModulus = (float) sqrt ( stNormal.fX * stNormal.fX + stNormal.fY * stNormal.fY + stNormal.fZ * stNormal.fZ );
      poPol->m_stPol.stNormal.fX = stNormal.fX / fModulus;
      poPol->m_stPol.stNormal.fY = stNormal.fY / fModulus;
      poPol->m_stPol.stNormal.fZ = stNormal.fZ / fModulus;

      // compute plan D component
      poPol->m_stPol.fD = poPol->m_stPol.stNormal.fX * poVert0->m_stVertex.nX +
                          poPol->m_stPol.stNormal.fY * poVert0->m_stVertex.nY +
                          poPol->m_stPol.stNormal.fZ * poVert0->m_stVertex.nZ;

      // next polygon
      m_oPolygons.mSelectNext ();
   }

   return TRUE;
}



/* Desalocate mesh data, if alocated
**********************************************************/
void CMesh::mDestroy ()
{
   // delete array of indexes (inside each polygon struct)
   CNodePolygon
      *poPol = (CNodePolygon*) m_oPolygons.mRetStart ();
   while ( poPol != NULL ) {

      // delete array of indexes
      delete poPol->m_stPol.padwVertexes;

      // next polygon
      poPol = (CNodePolygon*) poPol->m_poNext;
   }

   // remove data
   m_oPolygons.mRemoveAll ();
   m_oVertexes.mRemoveAll ();
   m_oPortals.mRemoveAll ();
}



/* Set polygons attributes to mesh.
Obs.: first add polygons then set their attributes, else you can
have errors because no polygon defined to set attribute.
Convension: TRUE = 1, FALSE = 0
bPortal - TRUE the polygon is a portal;
bCrossable - TRUE the polygon can be crossed, FALSE no (stop when collide)
dwTextureID - texture ID for the polygon;
nNdxPol1 - first polygon to apply attribute;
nNdxPol2 - last polygon to apply attribute;
Polygons that have indexes between nNdxPol1 and nNdxPol2
will have their attributes changed.
return TRUE - ok / FALSE - indexes out of range
Obs.: after change mesh data call CMesh::mBeginAddMesh.
**********************************************************/
BOOL CMesh::mSetPolygonsAttributes ( BOOL bPortal, BOOL bCrossable, DWORD dwTextureID, INT nNdxPol1, INT nNdxPol2 )
{
   // sort indexes
   INT
      nMin = nNdxPol1,
      nMax = nNdxPol2;
   if ( nNdxPol1 > nNdxPol2 ) {
      nMin = nNdxPol2,
      nMax = nNdxPol1;
   }

   // verify if polygons indexes in sector are out of range, if yes return
   if ( nMin < 0 || nMax >= (INT) m_oPolygons.mRetNumNodes () ) {
      printf ( MSG_ERROR_INDEX_OUT_OF_RANGE );
      return FALSE;
   }

   // set attribute values
   for ( INT ncI = nMin; ncI <= nMax; ncI ++ ) {
      ( (CNodePolygon*) m_oPolygons [ ncI ] )->m_stPol.bPortal = bPortal;
      ( (CNodePolygon*) m_oPolygons [ ncI ] )->m_stPol.bCrossable = bCrossable;
      ( (CNodePolygon*) m_oPolygons [ ncI ] )->m_stPol.dwTextureID = dwTextureID;
   }

   return TRUE;
}



/* Set polygons light intensity (level)
Convension: TRUE = 1, FALSE = 0
fLight can be between 0 and 1, is the light level (0 none, 1 maximum)
nNdxPol1 - first polygon to apply attribute;
nNdxPol2 - last polygon to apply attribute;
Polygons that have indexes between nNdxPol1 and nNdxPol2
will have their attributes changed.
return TRUE - ok / FALSE - indexes out of range
Obs.: after change mesh data call CMesh::mBeginAddMesh.
**********************************************************/
BOOL CMesh::mSetPolygonsLight ( float fLight, INT nNdxPol1, INT nNdxPol2 )
{
   // sort indexes
   INT
      nMin = nNdxPol1,
      nMax = nNdxPol2;
   if ( nNdxPol1 > nNdxPol2 ) {
      nMin = nNdxPol2,
      nMax = nNdxPol1;
   }

   // verify if polygons indexes in sector are out of range, if yes return
   if ( nMin < 0 || nMax >= (INT) m_oPolygons.mRetNumNodes () ) {
      printf ( MSG_ERROR_INDEX_OUT_OF_RANGE );
      return FALSE;
   }

   // set attribute value
   for ( INT ncI = nMin; ncI <= nMax; ncI ++ )
      ( (CNodePolygon*) m_oPolygons [ ncI ] )->m_stPol.fLight = fLight;

   return TRUE;
}



/* 1 - Add polygons
oIndexes is the list with vertex indexes that compose the polygon,
the indexes are acording vertex list in mesh.
return TRUE - ok \ FALSE - error, no memory
Obs.: after change mesh data call CMesh::mBeginAddMesh.
**********************************************************/
BOOL CMesh::mAddPolygon ( CList& oIndexes )
{
   // try create new polygon
   CNodePolygon
      *poPol = new CNodePolygon;
   if ( poPol == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // try alocate memory for polygon indexes
   poPol->m_stPol.padwVertexes = new DWORD [ oIndexes.mRetNumNodes () ];
   if ( poPol->m_stPol.padwVertexes == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      delete poPol;
      return FALSE;
   }

   // define polygon
   poPol->m_stPol.dwNumVertexes = oIndexes.mRetNumNodes ();
   // scan all list indexes and copy them to vertexes index array
   for ( DWORD dwcI = 0; dwcI < poPol->m_stPol.dwNumVertexes; dwcI ++ )
      poPol->m_stPol.padwVertexes [ dwcI ] = (DWORD) ( (CNodeInt*) oIndexes [ dwcI ] )->m_nValue;

   // define default attribute
   poPol->m_stPol.bPortal = FALSE;
   poPol->m_stPol.fLight = 1;
   poPol->m_stPol.dwTextureID = 0;

   // update mesh adding polygon and attribute
   m_oPolygons.mSetEnd ( (CNode*) poPol );

   return TRUE;
}



/* 2- Add polygons
Make a copy from the source polygon poSrc.
return TRUE - ok \ FALSE - error, no memory
Obs.: after change mesh data call CMesh::mBeginAddMesh.
**********************************************************/
BOOL CMesh::mAddPolygon ( CNodePolygon* poSrc )
{
   // try create new polygon
   CNodePolygon
      *poPol = new CNodePolygon;
   if ( poPol == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // define number of vertexes
   poPol->m_stPol.dwNumVertexes = poSrc->m_stPol.dwNumVertexes;

   // try alocate memory for polygon indexes
   poPol->m_stPol.padwVertexes = new DWORD [ poSrc->m_stPol.dwNumVertexes ];
   if ( poPol->m_stPol.padwVertexes == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      delete poPol;
      return FALSE;
   }

   // copy vertex indexes
   for ( DWORD dwcV = 0; dwcV < poPol->m_stPol.dwNumVertexes; dwcV ++ )
      poPol->m_stPol.padwVertexes [ dwcV ] = poSrc->m_stPol.padwVertexes [ dwcV ];

   // define attributes
   poPol->m_stPol.bPortal = poSrc->m_stPol.bPortal;
   poPol->m_stPol.fLight = poSrc->m_stPol.fLight;
   poPol->m_stPol.dwTextureID = poSrc->m_stPol.dwTextureID;
   poPol->m_stPol.bCrossable = poSrc->m_stPol.bCrossable;
   poPol->m_stPol.stNormal.fX = poSrc->m_stPol.stNormal.fX;
   poPol->m_stPol.stNormal.fY = poSrc->m_stPol.stNormal.fY;
   poPol->m_stPol.stNormal.fZ = poSrc->m_stPol.stNormal.fZ;
   poPol->m_stPol.fD = poSrc->m_stPol.fD;

   // update mesh adding polygon
   m_oPolygons.mSetEnd ( (CNode*) poPol );

   return TRUE;
}



/* Add vertexes coordenates.
bScale - TRUE - scale vertex (vertex * power(2,SCALEEXP))
return TRUE - ok \ FALSE - error, no memory
Obs.: after change mesh data call CMesh::mBeginAddMesh.
**********************************************************/
BOOL CMesh::mAddVertex ( INT nX, INT nY, INT nZ, BOOL bScale )
{
   // try create new vertex
   CNodeVertex
      *poVert = new CNodeVertex;
   if ( poVert == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // define vertex data
   // if need scale vertex
   if ( bScale ) {
      poVert->m_stVertex.nX = nX << SCALEEXP;  // == nX * power(2,SCALEEXP)
      poVert->m_stVertex.nY = nY << SCALEEXP;
      poVert->m_stVertex.nZ = nZ << SCALEEXP;
   }
   else {
      poVert->m_stVertex.nX = nX;
      poVert->m_stVertex.nY = nY;
      poVert->m_stVertex.nZ = nZ;
   }

   // update data
   m_oVertexes.mSetEnd ( (CNode*) poVert );

   return TRUE;
}



/* Add a portal polygon definition. between 2 sectors.
nNdxPol is the index, in polygon list, of the polygon that
is portal (the polygon list is in mesh data).
dwObjectID is the ID of destin object where sector is.
dwSectorID is the destin sector ID.
return TRUE - ok \ FALSE - error, no memory
Obs.: after change data call CSector::mBeginAddSector.
**********************************************************/
BOOL CMesh::mAddPortal ( DWORD dwNdxPol, DWORD dwObjectID, DWORD dwSectorID )
{
   // try create new portal node
   CNodePortal
      *poDestin = new CNodePortal;
   if ( poDestin == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // define portal data, destin sector
   poDestin->m_stPortal.dwObjectID = dwObjectID;
   poDestin->m_stPortal.dwSectorID = dwSectorID;

   // update sector adding portal
   m_oPortals.mAddKey ( dwNdxPol, (CNode*) poDestin );

   return TRUE;
}
